/*
 * DATlib   Copyright (c) 1995-1996 Marcus Meissner &
 *          Friedrich-Alexander Universitaet Erlangen-Nuernberg (FAU), IMMD4
 *          All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Marcus Meissner
 *    at the Friedrich-Alexander Universitaet Erlangen-Nuernberg (FAU), IMMD4
 *
 * 4. The name of the University or the author may not be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND THE AUTHOR ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR THE AUTHOR BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* $Id: if_userlevel.c,v 1.1 1997/03/16 20:29:34 msmeissn Exp $
 *
 * $Log: if_userlevel.c,v $
 * Revision 1.1  1997/03/16 20:29:34  msmeissn
 * Initial revision
 *
 *
 * if_userlevel.c: Interface routines to the underlying OS (unix for now)
 */

#ifndef _WIN32

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/fcntl.h>
#include <sys/sysmacros.h>	/* minor, used in mtio.h */
#include <sys/mtio.h>		/* used for MT_UNIT */
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#if !defined(_KERNEL) && !defined(getminor)
#define getminor(x) minor(x)
#endif
#include <sys/stat.h>

#include "dataudiolib.h"
#include "if_userlevel.h"
#include "mt_ext.h"

#ifdef sun
int open(char *fn,int mode,...);
#endif

/* the UNIX open interface */
int
da_downopen(char *fn,int opflags) {
	int	fd;

	fd=open(fn,opflags);
	if (	(fd==-1) &&
		(errno==ENOENT) &&
		(	(opflags&O_ACCMODE==O_RDWR)||
			(opflags&O_ACCMODE==O_WRONLY)
		)
	)
		fd=open(fn,opflags|O_CREAT,0666);
	if (fd==-1)
		perror("downopen:open:");
	return fd;
}

/* the init interface */
int
da_downinit(int fd,int opflags,char *fn,struct da_state *st) {
	struct	stat stbuf;
	struct	da_user_private *private;

	st->ifdata=(caddr_t)(private=(struct da_user_private*)malloc(sizeof(struct da_user_private)));
	if (-1==fstat(fd,&stbuf)) {
		perror("da_downopen:fstat");
		return -1;
	}
	private->fn=strdup(fn);
	private->mode = opflags;
	if (S_ISCHR(stbuf.st_mode)) {
		int	stateid;
		char	fnbuf[100];
		FILE	*F;

		private->type=DATAUDIO_DEV;
		private->major=major(stbuf.st_rdev);
		private->minor=minor(stbuf.st_rdev);
		stateid=(private->major<<16)|private->minor;
		sprintf(fnbuf,"/tmp/.dataudio_%d",stateid);
		if (NULL!=(F=fopen(fnbuf,"r"))){
			da_down_load_state(F,st);
  		  fclose(F);
                }
	} else
		private->type=DATAUDIO_FILE;
	return fd;
}

/* the UNIX close interface */
int
da_downclose(int fd) {
	char	fnbuf[100];
	FILE	*F;
	struct	da_state	*st;
	struct	da_user_private *private;
	int	stateid;
	extern	struct da_state *da_lookup(int id);
	
	st=da_lookup(fd);
	if (st==NULL) {fprintf(stderr,"da_lookup didn't find %d?\n",fd);exit(1);}
	private=(struct da_user_private*)st->ifdata;
	if (private->type==DATAUDIO_DEV) {
		stateid=(private->major<<16)|private->minor;
		sprintf(fnbuf,"/tmp/.dataudio_%d",stateid);
		if (NULL!=(F=fopen(fnbuf,"w"))){
			da_down_save_state(F,st);
		  fclose(F);
                }
	}
	return close(fd);
}

/* the UNIX write interface */
int
da_downwrite(int fd,unsigned char *buf,int size) {
	int	res,retries;

	res=-1;retries=20;
	while (retries-- && res==-1) {
		res=write(fd,buf,size);
	}
	return res;
}

/* the UNIX read interface */
int
da_downread(int fd,unsigned char *buf,int size) {
	int	ret,retries;

	retries=20;
	while (retries--) {
		ret=read(fd,buf,size);
		if (ret!=-1)
			break;
	}
	if (ret==-1)
		perror("read");
	return ret;
}

/* the UNIX ioctl interface, passes all necessary calls further down */
int 
da_downcontrol(int fd,int code,long arg) {
	union	mt_locate	mtloc;
	struct	da_locate	*daloc=(struct da_locate*)arg;
	struct	da_position	*dapos;
	unsigned	char	buf[20];
	int			res;
	struct da_state		*st;
	struct da_user_private	*priv;	

	if (code==DA_LOCATE) {
		st=da_lookup(fd);
		priv=(struct da_user_private*)st->ifdata;
		if (priv->mode & O_NDELAY)
			mtloc.locate_pno.immediate=1;
		else
			mtloc.locate_pno.immediate=0;
		switch (daloc->type) {
		case DA_LOC_PROGNR:
			mtloc.locate_pno.type=MT_LTYPE_PNO;
			mtloc.locate_pno.pno=l2bcd(daloc->u.prognr);
			break;
		case DA_LOC_PROGTIME:
			mtloc.locate_progtime.type=MT_LTYPE_PTIME;
			mtloc.locate_progtime.hour=l2bcd(daloc->u.time.hour);
			mtloc.locate_progtime.minute=l2bcd(daloc->u.time.minute);
			mtloc.locate_progtime.second=l2bcd(daloc->u.time.second);
			mtloc.locate_progtime.frame=l2bcd(daloc->u.time.frame);
			break;
		case DA_LOC_ABSTIME:
			mtloc.locate_abstime.type=MT_LTYPE_ATIME;
			mtloc.locate_abstime.hour=l2bcd(daloc->u.time.hour);
			mtloc.locate_abstime.minute=l2bcd(daloc->u.time.minute);
			mtloc.locate_abstime.second=l2bcd(daloc->u.time.second);
			mtloc.locate_abstime.frame=l2bcd(daloc->u.time.frame);
			break;
		case DA_LOC_RECTIME:
			mtloc.locate_runtime.type=MT_LTYPE_RTIME;
			mtloc.locate_runtime.hour=l2bcd(daloc->u.time.hour);
			mtloc.locate_runtime.minute=l2bcd(daloc->u.time.minute);
			mtloc.locate_runtime.second=l2bcd(daloc->u.time.second);
			mtloc.locate_runtime.frame=l2bcd(daloc->u.time.frame);
			break;
		}
		res=ioctl(fd,MTIOLOCATE,&mtloc);
		if (res==-1) {
			int	saveerrno=errno;
			perror("ioctl MTIOLOCATE");
			errno=saveerrno;
		}
		return res;
	}
	if (code==DA_READ_POSITION) {
		if (-1==ioctl(fd,MTIOREADPOS,buf)) {
			perror("ioctl MTIOREADPOS");
			return -1;
		}
		dapos=(struct da_position*)arg;
		dapos->prognr=buf[4]*256+buf[5]*16+buf[6];
		dapos->indexnr=buf[7];
		dapos->progtime.indexnr=dapos->indexnr;
		dapos->abstime.indexnr=dapos->indexnr;
		dapos->runtime.indexnr=dapos->indexnr;
		dapos->progtime.hour=buf[8];
		dapos->progtime.minute=buf[9];
		dapos->progtime.second=buf[10];
		dapos->progtime.frame=buf[11];
		dapos->abstime.hour=buf[12];
		dapos->abstime.minute=buf[13];
		dapos->abstime.second=buf[14];
		dapos->abstime.frame=buf[15];
		dapos->runtime.hour=buf[16];
		dapos->runtime.minute=buf[17];
		dapos->runtime.second=buf[18];
		dapos->runtime.frame=buf[19];
		return 0;
	}
	if (code==DA_REWIND) {
		struct	mtop	mtop;

		mtop.mt_op=MTREW;
		return ioctl(fd,MTIOCTOP,&mtop);
	}
#ifdef sun
	if (code==DA_ABORT)
		return ioctl(fd,MTIOABORT);
#endif
	if (code==DA_EJECT) {
		struct	mtop	mtop;

		mtop.mt_op=MTOFFL;
		return ioctl(fd,MTIOCTOP,&mtop);
	}
	return ioctl(fd,code,arg);
}

#endif /* def _WIN32 */